﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Contracts;
using System.Drawing;
using HeuristicLab.Common;
using System.Diagnostics;
using HeuristicLab.Optimization;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.JobBase;
using System.IO;
using HeuristicLab.Persistence.Default.Xml;
using HeuristicLab.PluginInfrastructure;
using System.Reflection;
using HeuristicLab.Hive.Experiment.Jobs;

namespace HeuristicLab.Hive.Experiment {
  public enum SnapshotRequestedState { Idle, Requested }

  [StorableClass]
  public class JobItem : Item {
    private static object locker = new object();

    public override Image ItemImage {
      get {
        if (jobDto.State == JobState.Offline) return HeuristicLab.Common.Resources.VS2008ImageLibrary.ExecutablePrepared;
        else if (jobDto.State == JobState.Calculating) return HeuristicLab.Common.Resources.VS2008ImageLibrary.ExecutableStarted;
        else if (jobDto.State == JobState.Aborted) return HeuristicLab.Common.Resources.VS2008ImageLibrary.ExecutableStopped;
        else if (jobDto.State == JobState.Failed) return HeuristicLab.Common.Resources.VS2008ImageLibrary.Error;
        else if (jobDto.State == JobState.Finished) return HeuristicLab.Common.Resources.VS2008ImageLibrary.ExecutableStopped;
        else return HeuristicLab.Common.Resources.VS2008ImageLibrary.Event;
      }
    }

    [Storable]
    private JobDto jobDto;
    public JobDto JobDto {
      get { return jobDto; }
      set {
        if (jobDto != value) {
          jobDto = value;
          OnJobDtoChanged();
          OnToStringChanged();
          OnItemImageChanged();
        }
      }
    }

    [Storable]
    private IJob job;
    public IJob Job {
      get { return job; }
      set {
        if (job != value) {
          job = value;
          IsFinalResultAvailable = job.ExecutionState == ExecutionState.Stopped;
          OnJobChanged();
        }
      }
    }

    [Storable]
    private ICollection<JobItem> childJobItems;
    public IEnumerable<JobItem> ChildJobItems {
      get { return childJobItems; }
    }

    [Storable]
    private bool isFinalResultAvailable;
    public bool IsFinalResultAvailable {
      get { return isFinalResultAvailable; }
      set {
        if (isFinalResultAvailable != value) {
          isFinalResultAvailable = value;
          if (isFinalResultAvailable)
            OnFinalResultAvailable();
        }
      }
    }
   
    public JobItem() {
      childJobItems = new List<JobItem>();
    }

    public JobItem(JobResult jobResult) : this() {
      this.jobDto = new JobDto();
      UpdateJob(jobResult);
    }

    public override string ToString() {
      if (jobDto != null) {
        return job.Name;
      } else {
        return base.ToString();
      }
    }

    public void UpdateJob(JobResult jobResult) {
      if (jobResult != null) {
        jobDto.Id = jobResult.Id;
        jobDto.DateCreated = jobResult.DateCreated;
        jobDto.DateCalculated = jobResult.DateCalculated;
        jobDto.DateFinished = jobResult.DateFinished;
        jobDto.Exception = jobResult.Exception;
        jobDto.Id = jobResult.Id;
        jobDto.Percentage = jobResult.Percentage;
        jobDto.State = jobResult.State;
        // what about parentJob
        OnJobStateChanged();
        OnToStringChanged();
        OnItemImageChanged();
      }
    }

    public void AddChildJob(JobItem child) {
      this.childJobItems.Add(child);
      OnChildJobAdded();
    }

    //public IEnumerable<JobItem> GetChildJobItems() {
    //  if (childJobItems == null)
    //    childJobItems = CreateChildJobItems();
    //  return childJobItems;
    //}

    //protected IEnumerable<JobItem> CreateChildJobItems() {
    //  return (from j in Job.ChildJobs
    //          select new JobItem() {
    //            Job = j,
    //            JobDto = new JobDto() {
    //              State = JobState.Offline
    //            }
    //          }).ToList();
    //}

    public SerializedJob ToSerializedJob() {
      MemoryStream memStream = new MemoryStream();
      XmlGenerator.Serialize(job, memStream);
      byte[] jobByteArray = memStream.ToArray();
      memStream.Dispose();

      UpdateRequiredPlugins();

      this.JobDto.CoresNeeded = job.CoresNeeded;
      this.JobDto.MemoryNeeded = job.MemoryNeeded;

      SerializedJob serializedJob = new SerializedJob() {
        JobInfo = jobDto,
        SerializedJobData = jobByteArray
      };

      return serializedJob;
    }

    private void UpdateRequiredPlugins() {
      // find out which which plugins are needed for the given object
      this.JobDto.PluginsNeeded = HivePluginInfoDto.FindPluginsNeeded(job.GetType());
    }
   
    public event EventHandler JobDtoChanged;
    private void OnJobDtoChanged() {
      LogMessage("JobDtoChanged");
      EventHandler handler = JobDtoChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler JobStateChanged;
    private void OnJobStateChanged() {
      LogMessage("JobStateChanged (State: " + this.JobDto.State + ", Percentage: " + this.JobDto.Percentage + ")");
      EventHandler handler = JobStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    
    public event EventHandler ChildJobAdded;
    private void OnChildJobAdded() {
      var handler = ChildJobAdded;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    
    public event EventHandler JobChanged;
    private void OnJobChanged() {
      var handler = JobChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler FinalResultAvailable;
    private void OnFinalResultAvailable() {
      var handler = FinalResultAvailable;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public void LogMessage(string message) {
      lock (locker) {
        if (job != null) {
          job.Log.LogMessage(message);
        }
      }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      LogMessage("I am beeing cloned");
      JobItem clone = (JobItem)base.Clone(cloner);
      clone.jobDto = (JobDto)cloner.Clone(this.jobDto);
      clone.job = (IJob)cloner.Clone(this.job);
      return clone;
    }

    public void UpdateFinalResults(IJob j) {
      if (this.Job is OptimizerJob && j is OptimizerJob) {
        ((OptimizerJob)this.Job).Optimizer.Runs.AddRange(((OptimizerJob)j).Optimizer.Runs);
        IsFinalResultAvailable = true; //job.ExecutionState == ExecutionState.Stopped; //??
      } else {
        throw new NotSupportedException();
      }
    }

    internal bool ReplaceChildJob(IJob job) {
      var matches = this.childJobItems.Where(child => child.Job.InternalId == job.InternalId);
      if (matches.Count() > 0) {
        matches.First().job = job;
        return true;
      }
      return false;
    }
  }
}
