﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Hive.Slave.Common;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.JobBase;
using HeuristicLab.Persistence.Core;
using HeuristicLab.Common;
using System.Xml.Serialization;
using HeuristicLab.Persistence.Default.Xml;
using System.IO;
using HeuristicLab.Core;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Hive.Slave.ExecutionEngine {
  public class Executor : MarshalByRefObject, IDisposable {
    public Guid JobId { get; set; }
    public IJob Job { get; set; }
    public MessageContainer.MessageType CurrentMessage { get; set; }
    public Exception CurrentException { get; set; }
    public MessageQueue Queue { get; set; }

    public bool JobIsFinished { get; set; }

    public ExecutionState ExecutionState {
      get {
        return Job.ExecutionState;
      }
    }

    public double Progress {
      get {
        return Job.Progress;
      }
    }

    public DateTime CreationTime { get; set; }

    public void Start(byte[] serializedJob) {
      CreationTime = DateTime.Now;
      Job = DeserializeJobObject(serializedJob);

      Job.JobStopped += new EventHandler(Job_JobStopped);
      Job.JobFailed += new EventHandler(Job_JobFailed);
      Job.NewChildJob += new EventHandler<EventArgs<IJob>>(Job_NewChildJob);
      Job.WaitForChildJobs += new EventHandler(Job_WaitForChildJobs);

      if (Job.ExecutionState == Core.ExecutionState.Paused) {
        Queue.AddMessage(new MessageContainerWithCallback<SerializedJobList>(MessageContainer.MessageType.GetChildJobs, JobId,
          (childjobs) => {
            Job.Resume(childjobs.Select(j => DeserializeJobObject(j.SerializedJobData)));
          }
        ));
      } else {
        Job.Prepare();
        Job.Start();
      }
    }

    void Job_NewChildJob(object sender, EventArgs<IJob> e) {
      byte[] jobByteArray = SerializeJobObject(e.Value);

      SerializedJob serializedJob = new SerializedJob() {
        JobInfo = new JobDto() {
          State = JobState.Offline,
          CoresNeeded = e.Value.CoresNeeded,
          MemoryNeeded = e.Value.MemoryNeeded,
          PluginsNeeded = HivePluginInfoDto.FindPluginsNeeded(e.Value.GetType()),
        },
        SerializedJobData = jobByteArray
      };

      Queue.AddMessage(new MessageContainerWithJob(MessageContainer.MessageType.AddChildJob, this.JobId, serializedJob));
    }

    void Job_WaitForChildJobs(object sender, EventArgs e) {
      // Pause the job and send it back to the hive. The server will awake it when all child-jobs are finished
      byte[] jobByteArray = SerializeJobObject(Job);

      SerializedJob serializedJob = new SerializedJob() {
        JobInfo = new JobDto() {
          Id = JobId,
          State = JobState.WaitForChildJobs
        },
        SerializedJobData = jobByteArray
      };

      JobIsFinished = true;
      Queue.AddMessage(new MessageContainerWithJob(MessageContainer.MessageType.PauseJob, this.JobId, serializedJob));
    }

    void Job_JobFailed(object sender, EventArgs e) {
      HeuristicLab.Common.EventArgs<Exception> ex = (HeuristicLab.Common.EventArgs<Exception>)e;
      CurrentException = ex.Value;
      Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.FinishedJob, JobId));
    }

    public void StartOnlyJob() {
      Job.Start();
    }

    public void Abort() {
      CurrentMessage = MessageContainer.MessageType.AbortJob;
      if (Job.ExecutionState != Core.ExecutionState.Stopped) {
        Job.Stop();
      }
    }

    void Job_JobStopped(object sender, EventArgs e) {
      if (CurrentMessage == MessageContainer.MessageType.NoMessage) {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.FinishedJob, JobId));
        JobIsFinished = true;
      } else if (CurrentMessage == MessageContainer.MessageType.RequestSnapshot) {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.SnapshotReady, JobId));
      } else if (CurrentMessage == MessageContainer.MessageType.AbortJob) {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.JobAborted, JobId));
      }
    }

    public byte[] GetSnapshot() {
      //if the job is still running, something went VERY bad.
      if (Job.ExecutionState == Core.ExecutionState.Started) {
        return null;
      } else {
        // Clear the Status message
        CurrentMessage = MessageContainer.MessageType.NoMessage;
        // Pack the whole job inside an xml document
        byte[] job = SerializeJobObject(Job);
        // Restart the job
        // Return the Snapshot
        return job;
      }
    }

    public byte[] GetFinishedJob() {
      //Job isn't finished!
      if (Job.ExecutionState == Core.ExecutionState.Started) {
        throw new InvalidStateException("Job is still running");
      } else {
        byte[] jobArr = SerializeJobObject(Job);
        return jobArr;
      }
    }

    public void RequestSnapshot() {
      CurrentMessage = MessageContainer.MessageType.RequestSnapshot;
      Job.Stop();
    }

    private byte[] SerializeJobObject(IJob job) {
      MemoryStream memStream = new MemoryStream();
      XmlGenerator.Serialize(job, memStream);
      byte[] jobByteArray = memStream.ToArray();
      memStream.Dispose();
      return jobByteArray;
    }

    private IJob DeserializeJobObject(byte[] sjob) {
      MemoryStream memStream = new MemoryStream(sjob);
      IJob job = XmlParser.Deserialize<IJob>(memStream);
      memStream.Dispose();
      return job;
    }

    private void RestoreJobObject(byte[] sjob) {
      Job = DeserializeJobObject(sjob);
    }

    public Executor() {
      CurrentMessage = MessageContainer.MessageType.NoMessage;
      JobIsFinished = false;
      //Job = new TestJob();
    }

    #region IDisposable Members

    public void Dispose() {
      Job.JobFailed -= new EventHandler(Job_JobFailed);
      Job.JobFailed -= new EventHandler(Job_JobStopped);
      Queue = null;
      Job = null;
    }

    #endregion
  }
}
